import sqlite3
import uuid
from os.path import join
from os import environ
from itertools import combinations
import pandas as pd
import numpy as np
from scipy.spatial import Delaunay

from aequilibrae.project.database_connection import database_connection
from aequilibrae.paths import Graph, TrafficClass, TrafficAssignment
from aequilibrae.matrix import AequilibraeMatrix
from aequilibrae.project.database_connection import ENVIRON_VAR

DELAUNAY_TABLE = 'delaunay_network'


class DelaunayAnalysis:
    def __init__(self):
        self.procedure_id = uuid.uuid4().hex

    def create_network(self, source='zones', overwrite=False):
        """Creates a delaunay network based on the existing model

        Args:
            *source* (:obj:`str`, `Optional`): Source of the centroids/zones. Defaults to *zones*, but can be *network*
            *overwrite path* (:obj:`bool`, `Optional`): Whether we should overwrite am existing Delaunay Network.
            Defaults to False
            """

        if source not in ['zones', 'network']:
            raise ValueError("Source must be 'zones' or 'network'")

        conn = database_connection()

        tables = pd.read_sql("SELECT name FROM sqlite_master WHERE type ='table'", conn)
        if DELAUNAY_TABLE in tables.name.values:
            if not overwrite:
                raise ValueError("Delaunay network already exist. Use the overwrite flag to re-run it")
            conn.execute(f'DROP TABLE {DELAUNAY_TABLE}')
            conn.execute('delete from geometry_columns where f_table_name=?', [DELAUNAY_TABLE])
            conn.commit()

        zone_sql = 'select zone_id node_id, X(st_centroid(geometry)) x, Y(st_centroid(geometry)) y from zones'
        network_sql = 'select node_id, X(geometry) x, Y(geometry) y from nodes where is_centroid=1'

        points = pd.read_sql(zone_sql, conn) if source == 'zones' else pd.read_sql(network_sql, conn)
        dpoints = np.array(points[['x', 'y']])
        all_edges = Delaunay(np.array(dpoints)).simplices
        edges = []
        for triangle in all_edges:
            links = list(combinations(triangle, 2))
            for i in links:
                f, t = sorted(list(i))
                edges.append([points.at[f, 'node_id'], points.at[t, 'node_id']])

        edges = pd.DataFrame(edges)
        edges.drop_duplicates(inplace=True)
        edges.columns = ['a_node', 'b_node']
        edges = edges.assign(direction=0, distance=0, link_id=np.arange(edges.shape[0]) + 1)
        edges = edges[['link_id', 'direction', 'a_node', 'b_node', 'distance']]
        edges.to_sql(DELAUNAY_TABLE, conn, index=False)

        # Now we create the geometries for the delaunay triangulation
        conn.execute(f"select AddGeometryColumn( '{DELAUNAY_TABLE}', 'geometry', 4326, 'LINESTRING', 'XY', 0);")
        conn.execute("CREATE UNIQUE INDEX unique_link_id_delaunay on delaunay_network(link_id)")

        node_geo_sql = '''INSERT INTO delaunay_network (link_id, geometry)
                             select lnk.link_id, MakeLine(nd.geometry, nf.geometry) from delaunay_network lnk
                                    inner join nodes nd on lnk.a_node=nd.node_id
                                    inner join nodes nf on lnk.b_node=nf.node_id
                           ON CONFLICT(link_id) DO UPDATE SET geometry=excluded.geometry'''

        zone_geo_sql = '''INSERT INTO delaunay_network (link_id, geometry)
                             select lnk.link_id, MakeLine(st_centroid(za.geometry), st_centroid(zb.geometry)) from delaunay_network lnk
                                    inner join zones za on lnk.a_node=za.zone_id
                                    inner join zones zb on lnk.b_node=zb.zone_id
                          ON CONFLICT(link_id) DO UPDATE SET geometry=excluded.geometry;'''

        sql = zone_geo_sql if source == 'zones' else node_geo_sql
        conn.execute(sql)
        # Updates link distance
        conn.execute('update delaunay_network set distance=GeodesicLength(geometry);')
        conn.commit()
        conn.close()

    def assign_matrix(self, matrix: AequilibraeMatrix, result_name: str):
        conn = database_connection()

        sql = f"select link_id, direction, a_node, b_node, distance, 1 capacity from {DELAUNAY_TABLE}"

        df = pd.read_sql(sql, conn)
        centroids = np.array(np.unique(np.hstack((df.a_node.values, df.b_node.values))), int)

        g = Graph()
        g.mode = 'delaunay'
        g.network = df
        g.prepare_graph(centroids)
        g.set_blocked_centroid_flows(True)

        tc = TrafficClass('delaunay', g, matrix)
        ta = TrafficAssignment()
        ta.set_classes([tc])
        ta.set_time_field('distance')
        ta.set_capacity_field('capacity')
        ta.set_vdf('BPR')
        ta.set_vdf_parameters({"alpha": 0, "beta": 1.0})
        ta.set_algorithm('all-or-nothing')
        ta.execute()

        report = {"setup": str(ta.info())}
        data = [result_name, "Delaunay assignment", self.procedure_id, str(report), ta.procedure_date, '']
        conn.execute("""Insert into results(table_name, procedure, procedure_id, procedure_report, timestamp,
                                            description) Values(?,?,?,?,?,?)""", data)
        conn.commit()
        conn.close()

        cols = []
        for x in matrix.view_names:
            cols.extend([f'{x}_ab', f'{x}_ba', f'{x}_tot'])
        df = ta.results()[cols]
        conn = sqlite3.connect(join(environ[ENVIRON_VAR], "results_database.sqlite"))
        df.to_sql(result_name, conn)
        conn.close()



import os
import yaml
from warnings import warn
from copy import deepcopy


class Parameters:
    """
    Global parameters module

    Parameters are used in many procedures, and are often defined only in the parameters.yml file ONLY
    Parameters are organized in the following groups:

    * assignment
    * distribution

    * system
        - cpus: Maximum threads to be used in any procedure

        - default_directory: If is the directory QGIS file opening/saving dialogs will try to open as standard

        - driving side: For purposes of plotting on QGIS

        - logging: Level of logging to be written to temp/aequilibrae.log: Levels are those from the Python logging library
                    - 0: 'NOTSET'
                    - 10: 'DEBUG'
                    - 20: 'INFO'
                    - 30: 'WARNING'
                    - 40: 'ERROR'
                    - 50: 'CRITICAL'
    * report zeros
    * temp directory
    """

    def __init__(self):
        """ Loads parameters from file. The place is always the same. The root of the package"""

        proj_path = os.environ.get('AEQUILIBRAE_PROJECT_PATH', '')
        default_path = os.path.dirname(os.path.realpath(__file__))

        self.file = os.path.join(proj_path, "parameters.yml")
        self.file_default = os.path.join(default_path, "parameters.yml")

        with open(self.file_default, "r") as yml:
            self._default = yaml.load(yml, Loader=yaml.SafeLoader)

        if os.path.isfile(self.file):
            with open(self.file, "r") as yml:
                self.parameters = yaml.load(yml, Loader=yaml.SafeLoader)
        else:
            self.parameters = deepcopy(self._default)
            if proj_path:
                warn('No pre-existing parameter file exists for this project. Will use default')

    def write_back(self):
        """Writes the parameters back to file"""
        with open(self.file, "w") as stream:
            yaml.dump(self.parameters, stream, default_flow_style=False)

    def restore_default(self):
        """Restores parameters to generic default"""
        self.parameters = self._default
        self.write_back()



from os.path import join


class Log:
    """API entry point to the log file contents

    ::

        from aequilibrae import Project

        p = Project()
        p.open('path/to/project/folder')

        log = p.log()

        # We get all entries for the log file
        entries = log.contents()

        # Or clear everything (NO UN-DOs)
        log.clear()
    """

    def __init__(self, project_base_path: str):
        self.log_file_path = join(project_base_path, "aequilibrae.log")

    def contents(self) -> list:
        """Returns contents of log file

        Return:
            *log_contents* (:obj:`list`): List with all entries in the log file
        """

        with open(self.log_file_path, 'r') as file:
            return [x.strip() for x in file.readlines()]

    def clear(self):
        """Clears the log file. Use it wisely"""
        with open(self.log_file_path, 'w') as _:
            pass



from unittest import TestCase, TestLoader

from sqlite3 import IntegrityError
from copy import copy, deepcopy
import os
from shutil import copytree, rmtree
from random import randint, random
import uuid
from tempfile import gettempdir
from shapely.geometry import Point
import shapely.wkb
from aequilibrae.project import Project

from ...data import siouxfalls_project

TestLoader.sortTestMethodsUsing = None


class TestNodes(TestCase):
    def setUp(self) -> None:
        os.environ['PATH'] = os.path.join(gettempdir(), 'temp_data') + ';' + os.environ['PATH']

        self.proj_dir = os.path.join(gettempdir(), uuid.uuid4().hex)
        copytree(siouxfalls_project, self.proj_dir)

        self.project = Project()
        self.project.open(self.proj_dir)
        self.network = self.project.network
        self.curr = self.project.conn.cursor()

    def tearDown(self) -> None:
        self.project.close()
        del self.curr
        try:
            rmtree(self.proj_dir)
        except Exception as e:
            print(f'Failed to remove at {e.args}')

    def test_get(self):
        nodes = self.network.nodes
        nd = randint(1, 24)
        node = nodes.get(nd)

        self.assertEqual(node.node_id, nd, 'get node returned wrong object')

        # Make sure that if we renumber itg we would not get it again
        node.renumber(200)
        with self.assertRaises(ValueError):
            node = nodes.get(nd)

    def test_save(self):
        nodes = self.network.nodes
        chosen = [randint(1, 24) for _ in range(5)]
        while len(chosen) != len(set(chosen)):
            chosen = [randint(1, 24) for _ in range(5)]
        coords = []
        for nd in chosen:
            node = nodes.get(nd)
            node.is_centroid = 0
            x = node.geometry.x + random()
            y = node.geometry.y + random()
            coords.append([x, y])
            node.geometry = Point([x, y])

        nodes.save()
        for nd, coords in zip(chosen, coords):
            x, y = coords
            self.curr.execute('Select is_centroid, asBinary(geometry) from nodes where node_id=?;', [nd])
            flag, wkb = self.curr.fetchone()
            self.assertEqual(flag, 0, 'Saving of is_centroid failed')

            geo = shapely.wkb.loads(wkb)
            self.assertEqual(geo.x, x, 'Geometry X saved wrong')
            self.assertEqual(geo.y, y, 'Geometry Y saved wrong')

    def test_fields(self):
        nodes = self.network.nodes
        f_editor = nodes.fields()

        fields = sorted(f_editor.all_fields())
        self.curr.execute('pragma table_info(nodes)')
        dt = self.curr.fetchall()

        actual_fields = set([x[1] for x in dt if x[1] != 'ogc_fid'])
        actual_fields = sorted(list(actual_fields))

        self.assertEqual(fields, actual_fields, 'Table editor is weird for table nodes')

    def test_copy(self):
        nodes = self.network.nodes
        with self.assertRaises(Exception):
            _ = copy(nodes)
        with self.assertRaises(Exception):
            _ = deepcopy(nodes)

    def test_new_centroid(self):
        nodes = self.network.nodes

        with self.assertRaises(Exception):
            node = nodes.new_centroid(1)

        tot_prev_centr = self.network.count_centroids()
        tot_prev_nodes = self.network.count_nodes()
        node = nodes.new_centroid(100)
        self.assertEqual(1, node.is_centroid, 'Creating new centroid returned wrong is_centroid value')
        node.geometry = Point(1, 1)
        node.save()

        self.assertEqual(tot_prev_centr + 1, self.network.count_centroids(), 'Failed to add centroids')
        self.assertEqual(tot_prev_nodes + 1, self.network.count_nodes(), 'Failed to add centroids')



from tempfile import gettempdir
import urllib.request
import platform
import zipfile
from os.path import join
from os import walk

pth = 'https://github.com/AequilibraE/aequilibrae/releases/download/V0.6.0.post1/mod_spatialite-NG-win-amd64.zip'

outfolder = gettempdir()

dest_path = join(outfolder, "mod_spatialite-NG-win-amd64.zip")
urllib.request.urlretrieve(pth, dest_path)

fldr = join(outfolder, 'temp_data')
zipfile.ZipFile(dest_path).extractall(fldr)

if 'WINDOWS' in platform.platform().upper():
    # We now set sqlite. Only needed in thge windows server in Github
    plats = {'x86': 'https://sqlite.org/2020/sqlite-dll-win32-x86-3320100.zip',
             'x64': 'https://sqlite.org/2020/sqlite-dll-win64-x64-3320100.zip'}

    outfolder = 'C:/'
    zip_path64 = join(outfolder, "sqlite-dll-win64-x64-3320100.zip")
    urllib.request.urlretrieve(plats['x64'], zip_path64)

    zip_path86 = join(outfolder, "sqlite-dll-win32-x86-3320100.zip")
    urllib.request.urlretrieve(plats['x86'], zip_path86)

    root = 'C:/hostedtoolcache/windows/Python/'
    file = 'sqlite3.dll'
    for d, subD, f in walk(root):
        if file in f:
            if 'x64' in d:
                zipfile.ZipFile(zip_path64).extractall(d)
            else:
                zipfile.ZipFile(zip_path86).extractall(d)
            print(f'Replaces {d}')



import os
from uuid import uuid4
from random import choice
from shutil import copytree
from os.path import join
from tempfile import gettempdir
from unittest import TestCase
from aequilibrae.project.field_editor import FieldEditor, ALLOWED_CHARACTERS
from aequilibrae import Project
from ...data import siouxfalls_project


class TestFieldEditor(TestCase):
    my_tables = ["link_types", "links", "modes", "nodes"]

    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(gettempdir(), "temp_data") + ";" + os.environ["PATH"]
        self.temp_proj_folder = join(gettempdir(), uuid4().hex)
        copytree(siouxfalls_project, self.temp_proj_folder)
        self.proj = Project()
        self.proj.open(self.temp_proj_folder)

    def tearDown(self) -> None:
        self.proj.close()

    def randomword(self, length):
        val = "".join(choice(ALLOWED_CHARACTERS) for i in range(length))
        if val[0] == "_" or val[-1] == "_":
            return self.randomword(length)
        return val

    def test_building(self):
        for tab in ["modes", "links", "nodes", "link_types"]:
            table = FieldEditor(tab)
            qry = f'select count(*) from "attributes_documentation" where name_table="{tab}"'
            q = self.proj.conn.execute(qry).fetchone()[0]
            self.assertEqual(q, len(table._original_values), "Meta table populated with the wrong number of elements")

    def test_add(self):
        for tab in self.my_tables:
            table = FieldEditor(tab)
            qry = f'select count(*) from "attributes_documentation" where name_table="{tab}"'
            q = self.proj.conn.execute(qry).fetchone()[0]
            one = choice(list(table._original_values.keys()))
            self.proj.conn.commit()
            with self.assertRaises(ValueError) as em:
                table.add(one, self.randomword(30))
            self.assertEqual("attribute_name already exists", str(em.exception), "failed in the wrong place")

            with self.assertRaises(ValueError) as em:
                table.add(f"{self.randomword(5)} {5}", self.randomword(30))
            self.assertEqual(
                'attribute_name can only contain letters, numbers and "_"',
                str(em.exception),
                "failed in the wrong place",
            )

            with self.assertRaises(ValueError):
                table.add(choice("0123456789") + self.randomword(20), self.randomword(30))
            new_one = self.randomword(20)
            while new_one[0] in "0123456789":
                new_one = self.randomword(20)
            table.add(new_one, self.randomword(30))
            self.proj.conn.commit()
            curr = self.proj.conn.cursor()
            curr.execute(f'select count(*) from "attributes_documentation" where name_table="{tab}"')
            q2 = curr.fetchone()[0]
            self.assertEqual(q + 1, q2, "Adding element did not work")

            # If query fails, we failed to add new field to the database
            curr.execute(f'select "{new_one}" from "attributes_documentation" where name_table="{tab}"')

            if "alpha" in table._original_values.keys():
                self.assertEqual(table.alpha, "Available for user convenience", "not being able to retrieve values")

            self.proj.conn.commit()
            del curr

    def test_check_completeness(self):
        for table in self.my_tables:
            curr = self.proj.conn.cursor()

            # We add a bogus record to the attribute list
            val = self.randomword(30).lower()
            qry = 'INSERT INTO "attributes_documentation" VALUES (?,?," ");'
            curr.execute(qry, (table, val))
            self.proj.conn.commit()

            curr.execute(f'Select name_table from "attributes_documentation" where attribute="{val}"')
            self.assertEqual(curr.fetchone()[0], table, "Failed to insert bogus value")

            # Then we add a new field to the table
            val2 = self.randomword(10)
            curr.execute(f'Alter table "{table}" add column "{val2}" NUMERIC;')
            curr.execute(f"pragma table_info({table})")
            fields = [x[1] for x in curr.fetchall() if x[1] == val2]
            self.assertEqual([val2], fields, "failed to add a new field")

            table = FieldEditor(table)
            self.proj.conn.commit()
            curr = self.proj.conn.cursor()
            curr.execute(f'Select count(*) from "attributes_documentation" where attribute="{val}"')
            self.assertEqual(curr.fetchone()[0], 0, f"clean the table on loading failed {val}")

            curr.execute(f'Select count(*) from "attributes_documentation" where attribute="{val2}"')
            self.assertEqual(curr.fetchone()[0], 1, "clean the table on loading failed")
            print(table._original_values[val2])

            self.proj.conn.commit()
            del curr

    def test_save(self):
        for tab in ["modes", "links", "nodes", "link_types"]:
            table = FieldEditor(tab)
            random_val = self.randomword(30)
            if "alpha" in table._original_values.keys():
                table.alpha = random_val
                table.save()
                table2 = FieldEditor(tab)

                self.assertEqual(table2.alpha, random_val, "Did not save values properly")

            if "link_id" in table._original_values.keys():
                table.link_id = random_val
                table.save()
                table2 = FieldEditor(tab)

                self.assertEqual(table2.link_id, random_val, "Did not save values properly")

            if "node_id" in table._original_values.keys():
                table.node_id = random_val
                table.save()
                table2 = FieldEditor(tab)

                self.assertEqual(table2.node_id, random_val, "Did not save values properly")

            self.proj.conn.commit()



import os
import sys
import tempfile
import logging
from .parameters import Parameters
import glob

sys.dont_write_bytecode = True


# TODO: Add tests for logging
def StartsLogging():
    # CREATE THE LOGGER
    p = Parameters()
    par = p.parameters
    if p.parameters is None:
        par = p._default
    temp_folder = par["system"]["logging_directory"]
    do_log = par["system"]["logging"]
    if not os.path.isdir(temp_folder):
        temp_folder = tempfile.gettempdir()

    if do_log:
        log_file = os.path.join(temp_folder, "aequilibrae.log")
        if not os.path.isfile(log_file):
            a = open(log_file, "w")
            a.close()

    logger = logging.getLogger("aequilibrae")
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(asctime)s;%(name)s;%(levelname)s ; %(message)s")

    if not len(logger.handlers):
        if do_log:
            ch = logging.FileHandler(log_file)
        ch.setFormatter(formatter)
        ch.name = 'aequilibrae'
        ch.setLevel(logging.DEBUG)
        logger.addHandler(ch)
    return logger


def cleaning():
    p = tempfile.gettempdir() + "/aequilibrae_*"
    for f in glob.glob(p):
        try:
            os.unlink(f)
        except Exception as err:
            logger.warning(err.__str__())


logger = StartsLogging()
cleaning()



# from unittest import TestCase
# from aequilibrae.project import Project as AequilibraEProject
# from ...data import project_file
#
#
# class TestAequilibraEProject(TestCase):
#     def load_model(self):
#         self.model = AequilibraEProject()
#         self.model.load(project_file)
#         self.model.load_model(project_file)
#
#     # def test_load_model(self):
#     #     model = AequilibraEProject()
#     #     model.load_model(project_file)
#
#     def test_metadata(self):
#         self.load_model()
#         assert self.model.metadata("model_name") == "My model test project"
#         assert self.model.metadata("author") == "Pedro Camargo"
#         assert self.model.metadata("contact_info") == "c@margo.co"
#         assert self.model.metadata("description") == "Project generic description"
#         assert self.model.metadata("origin_of_network") == "OSM"
#         assert self.model.metadata("origin_of_demand") == "Custom"
#         assert self.model.metadata("license") == "Same as AequilibraE"
#
#         # try:
#         #     x = self.model.metadata("xxxx")
#         #     self.fail("Failed to return an error on {} querying metadata  for AequilibraE project")
#         # except ValueError:
#         #     pass
#
#     def test_write_metadata(self):
#         self.load_model()
#         self.model.write_metadata("model_name", "Did it work?")
#         assert self.model.metadata("model_name") == "Did it work?"
#
#         self.model.write_metadata("model_name", "My model test project")



from unittest import TestCase
import sqlite3
import tempfile
import random
import os
import uuid
from shutil import copytree, rmtree
from aequilibrae import Project
import string
from ...data import siouxfalls_project


class TestAbout(TestCase):
    def setUp(self) -> None:
        os.environ['PATH'] = os.path.join(tempfile.gettempdir(), 'temp_data') + ';' + os.environ['PATH']
        self.temp_proj_folder = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
        copytree(siouxfalls_project, self.temp_proj_folder)
        self.proj = Project()
        self.proj.open(self.temp_proj_folder)
        self.curr = self.proj.conn.cursor()

    def tearDown(self) -> None:
        self.proj.close()

    def test_create_and_list(self):
        self.proj.about.create()

        with self.assertWarns(expected_warning=Warning):
            self.proj.about.create()

        list = self.proj.about.list_fields()
        expected = ['model_name', 'region', 'description', 'author', 'license', 'scenario_name', 'year',
                    'scenario_description', 'model_version', 'project_id', 'aequilibrae_version', 'projection']

        failed = set(list).symmetric_difference(set(expected))
        if failed:
            self.fail('About table does not have all expected fields')

    # idea from https://stackoverflow.com/a/2030081/1480643
    def randomword(self, length):
        letters = string.ascii_lowercase + '_'
        return ''.join(random.choice(letters) for i in range(length))

    def test_add_info_field(self):
        self.proj.about.create()

        all_added = set()
        for t in range(30):
            k = self.randomword(random.randint(1, 15))
            if k not in all_added:
                all_added.add(k)
                self.proj.about.add_info_field(k)

        curr = self.proj.conn.cursor()
        curr.execute("select infoname from 'about'")

        charac = [x[0] for x in curr.fetchall()]
        for k in all_added:
            if k not in charac:
                self.fail(f'Failed to add {k}')

        # Should fail when trying to add a repeated guy
        with self.assertRaises(sqlite3.IntegrityError):
            self.proj.about.add_info_field('description')

        # Should fail when trying to add a repeated guy
        with self.assertRaises(ValueError):
            self.proj.about.add_info_field('descr1ption')

    def test_write_back(self):
        self.proj.about.create()
        self.proj.about.add_info_field('good_info_field_perhaps')

        val = self.randomword(random.randint(1, 15))
        self.proj.about.good_info_field_perhaps = val

        val2 = self.randomword(random.randint(30, 250))
        self.proj.about.description = val2

        self.proj.about.write_back()

        self.proj.close()
        del self.proj

        self.proj = Project()
        self.proj.open(self.temp_proj_folder)
        self.assertEqual(val, self.proj.about.good_info_field_perhaps, 'failed to save data to about table')
        self.assertEqual(val2, self.proj.about.description, 'failed to save data to about table')



import sqlite3
import os
import shutil
from aequilibrae.starts_logging import StartsLogging
from aequilibrae.project.network import Network
from aequilibrae.project.zoning import Zoning
from aequilibrae.project.data import Matrices
from aequilibrae.project.about import About
from aequilibrae.project.database_connection import database_connection, ENVIRON_VAR
from aequilibrae.parameters import Parameters
from aequilibrae.log import Log
import warnings
from aequilibrae import logger
from aequilibrae.reference_files import spatialite_database
from .project_creation import initialize_tables
import logging
from .project_cleaning import clean


class Project:
    """AequilibraE project class

    ::

        from aequilibrae.project import Project

        existing = Project()
        existing.load('path/to/existing/project/folder')

        newfile = Project()
        newfile.new('path/to/new/project/folder')
        """

    def __init__(self):
        self.path_to_file: str = None
        self.project_base_path = ""
        self.source: str = None
        self.conn: sqlite3.Connection = None
        self.network: Network = None
        self.about: About = None
        self.logger: logging.Logger = None

    def open(self, project_path: str) -> None:
        """
        Loads project from disk

        Args:
            *project_path* (:obj:`str`): Full path to the project data folder. If the project inside does
            not exist, it will fail.
        """

        if self.__other_project_still_open():
            raise Exception("You already have a project open. Close that project before opening another one")

        file_name = os.path.join(project_path, "project_database.sqlite")
        if not os.path.isfile(file_name):
            raise FileNotFoundError("Model does not exist. Check your path and try again")

        self.project_base_path = project_path
        self.path_to_file = file_name
        self.source = self.path_to_file
        os.environ[ENVIRON_VAR] = self.project_base_path
        self.conn = database_connection()

        self.__load_objects()
        self.__set_logging_path()
        logger.info(f"Opened project on {self.project_base_path}")
        self.logger = logger
        clean()

    def new(self, project_path: str) -> None:
        """Creates a new project

        Args:
            *project_path* (:obj:`str`): Full path to the project data folder. If folder exists, it will fail
        """
        if self.__other_project_still_open():
            raise Exception("You already have a project open. Close that project before creating a new one")

        self.project_base_path = project_path
        self.path_to_file = os.path.join(self.project_base_path, "project_database.sqlite")
        self.source = self.path_to_file

        if os.path.isdir(project_path):
            raise FileNotFoundError("Location already exists. Choose a different name or remove the existing directory")
        os.environ[ENVIRON_VAR] = self.project_base_path

        self.__create_empty_project()
        self.__load_objects()
        self.about.create()
        self.__set_logging_path()
        self.logger = logger
        logger.info(f"Created project on {self.project_base_path}")

    def close(self) -> None:
        """Safely closes the project"""
        if ENVIRON_VAR in os.environ:
            self.conn.commit()
            clean()
            self.conn.close()
            for obj in [self.parameters, self.network]:
                del obj
            del os.environ[ENVIRON_VAR]
            del self.network.link_types
            del self.network.modes
            logger.info(f"Closed project on {self.project_base_path}")
        else:
            warnings.warn("There is no Aequilibrae project open that you could close")

    def load(self, project_path: str) -> None:
        """
        Loads project from disk

        Args:
            *project_path* (:obj:`str`): Full path to the project data folder. If the project inside does
            not exist, it will fail.
        """
        warnings.warn(f"Function has been deprecated. Use my_project.open({project_path}) instead", DeprecationWarning)
        self.open(project_path)

    def log(self) -> Log:
        """Returns a log object

        allows the user to read the log or clear it"""
        return Log(self.project_base_path)

    def __load_objects(self):
        matrix_folder = os.path.join(self.project_base_path, "matrices")
        if not os.path.isdir(matrix_folder):
            os.mkdir(matrix_folder)

        self.network = Network(self)
        self.about = About(self.conn)

    @property
    def matrices(self) -> Matrices:
        return Matrices()

    @property
    def parameters(self) -> dict:
        return Parameters().parameters

    def check_file_indices(self) -> None:
        """ Makes results_database.sqlite and the matrices folder compatible with project database
        """
        raise NotImplementedError

    @property
    def zoning(self):
        return Zoning(self.network)

    def __create_empty_project(self):

        # We create the project folder and create the base file
        os.mkdir(self.project_base_path)
        shutil.copyfile(spatialite_database, self.path_to_file)

        self.conn = database_connection()

        # Write parameters to the project folder
        p = Parameters()
        p.parameters["system"]["logging_directory"] = self.project_base_path
        p.write_back()
        _ = StartsLogging()

        # Create actual tables
        cursor = self.conn.cursor()
        cursor.execute("PRAGMA foreign_keys = ON;")
        self.conn.commit()
        initialize_tables(self.conn)

    def __other_project_still_open(self) -> bool:
        if ENVIRON_VAR in os.environ:
            return True
        return False

    def __set_logging_path(self):
        p = Parameters()
        par = p.parameters
        if p.parameters is None:
            par = p._default
        do_log = par["system"]["logging"]
        for handler in logger.handlers:
            if handler.name == "aequilibrae":
                logger.removeHandler(handler)
        if do_log:
            formatter = logging.Formatter("%(asctime)s;%(name)s;%(levelname)s ; %(message)s")
            log_file = os.path.join(self.project_base_path, "aequilibrae.log")
            if not os.path.isfile(log_file):
                a = open(log_file, "w")
                a.close()
            ch = logging.FileHandler(log_file)
            ch.name = "aequilibrae"
            ch.setFormatter(formatter)
            ch.setLevel(logging.DEBUG)
            logger.addHandler(ch)



from aequilibrae.starts_logging import logger
from aequilibrae.parameters import Parameters
from aequilibrae.project.data import Matrices
from aequilibrae.log import Log
from aequilibrae import distribution
from aequilibrae import matrix
from aequilibrae import transit
from aequilibrae import project
import warnings

compiled = True
try:
    from aequilibrae.paths.AoN import path_computation
except Exception as e:
    compiled = False
    warnings.warn(f"Failed to import compiled modules. {e.args}")

if compiled:
    from aequilibrae.distribution import Ipf, GravityApplication, GravityCalibration, SyntheticGravityModel
    from aequilibrae.matrix import AequilibraeMatrix, AequilibraeData

    from aequilibrae.paths.network_skimming import NetworkSkimming
    from aequilibrae.paths.traffic_class import TrafficClass
    from aequilibrae.paths.vdf import VDF
    from aequilibrae.paths.all_or_nothing import allOrNothing
    from aequilibrae.paths.traffic_assignment import TrafficAssignment
    from aequilibrae.paths.graph import Graph
    from aequilibrae.paths.assignment_paths import AssignmentPaths
    from aequilibrae.project import Project
    from aequilibrae.transit.gtfs import GTFS, create_gtfsdb
    from aequilibrae.paths.results import AssignmentResults, SkimResults, PathResults
    from aequilibrae.paths import release_version as __version__

    from aequilibrae import paths
name = "aequilibrae"



import uuid
import os
import tempfile
from unittest import TestCase
from aequilibrae.project.database_connection import database_connection
from aequilibrae.project import Project


class TestDatabaseConnection(TestCase):
    def setUp(self) -> None:
        os.environ['PATH'] = os.path.join(tempfile.gettempdir(), 'temp_data') + ';' + os.environ['PATH']

    def test_database_connection(self):
        # Errors when project does not exist
        with self.assertRaises(FileExistsError):
            _ = database_connection()

    def test_connection_with_new_project(self):
        temp_proj_folder = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
        proj = Project()
        proj.new(temp_proj_folder)
        proj.close()

        proj = Project()
        proj.open(temp_proj_folder)
        conn = database_connection()
        cursor = conn.cursor()
        cursor.execute('select count(*) from links')

        self.assertEqual(cursor.fetchone()[0], 0, "Returned more links thant it should have")
        proj.close()



from os.path import join
from tempfile import gettempdir
from uuid import uuid4
from unittest import TestCase
from aequilibrae.utils.create_example import create_example
from aequilibrae import Parameters


class Test(TestCase):
    def test_create_example(self):
        par = Parameters()._default
        for p in ["nauru", "sioux_falls"]:
            pth = join(gettempdir(), f"proj_example_{p}_{uuid4().hex}")
            proj = create_example(pth, p)
            parproj = proj.parameters
            self.assertEqual(parproj.keys(), par.keys(), f"Wrong parameter keys for {p} example")
            proj.close()



import zipfile
import os
from os.path import dirname, join
from aequilibrae import Project


def create_example(path: str, from_model='sioux_falls') -> Project:
    """Copies an example model to a new project project and returns the project handle

    Args:
        *path* (:obj:`str`): Path where to create a new model. must be a non-existing folder/directory.
        *from_model path* (:obj:`str`, `Optional`): Example to create from *sioux_falls* or *nauru*. Defaults to
        *sioux_falls*
    Returns:
        *project* (:obj:`Project`): Aequilibrae Project handle (open)

        """
    if os.path.isdir(path):
        raise FileExistsError('Cannot overwrite an existing directory')

    if not os.path.isfile(join(dirname(__file__), f'../reference_files/{from_model}.zip')):
        raise FileExistsError('Example not found')

    os.mkdir(path)
    zipfile.ZipFile(join(dirname(__file__), f'../reference_files/{from_model}.zip')).extractall(path)
    proj = Project()
    proj.open(path)
    return proj



import sys
from os.path import join, isfile
import threading
import importlib.util as iutil
from uuid import uuid4
from multiprocessing.dummy import Pool as ThreadPool
from datetime import datetime
from aequilibrae.project.data import Matrices
from aequilibrae.paths.multi_threaded_skimming import MultiThreadedNetworkSkimming
from aequilibrae.paths.results.skim_results import SkimResults
from aequilibrae.utils import WorkerThread
from aequilibrae import logger

try:
    from aequilibrae.paths.AoN import skimming_single_origin
except ImportError as ie:
    logger.warning(f"Could not import procedures from the binary. {ie.args}")

spec = iutil.find_spec("PyQt5")
pyqt = spec is not None
if pyqt:
    from PyQt5.QtCore import pyqtSignal

spec = iutil.find_spec("openmatrix")
has_omx = spec is not None

sys.dont_write_bytecode = True


class NetworkSkimming(WorkerThread):
    """

    ::

        from aequilibrae.paths.network_skimming import NetworkSkimming
        from aequilibrae.project import Project

        project = Project()
        project.open(self.proj_dir)
        network = self.project.network

        network.build_graphs()
        graph = network.graphs['c']
        graph.set_graph(cost_field="distance")
        graph.set_skimming("distance")

        skm = NetworkSkimming(graph)
        skm.execute()

        # The skim report (if any error generated) is available here
        skm.report

        # To access the skim matrix directly from its temporary file
        matrix = skm.results.skims

        # Or you can save the results to disk
        skm.save_to_project('skimming result')

        # Or specify the AequilibraE's matrix file format
        skm.save_to_project('skimming result', 'aem')

        project.close()
    """

    if pyqt:
        skimming = pyqtSignal(object)

    def __init__(self, graph, origins=None):
        WorkerThread.__init__(self, None)

        self.origins = origins
        self.graph = graph
        self.results = SkimResults()
        self.aux_res = MultiThreadedNetworkSkimming()
        self.report = []
        self.procedure_id = ""
        self.procedure_date = ""
        self.cumulative = 0

    def doWork(self):
        self.execute()

    def execute(self):
        """Runs the skimming process as specified in the graph"""
        if pyqt:
            self.skimming.emit(["zones finalized", 0])

        self.results.prepare(self.graph)
        self.aux_res = MultiThreadedNetworkSkimming()
        self.aux_res.prepare(self.graph, self.results)

        pool = ThreadPool(self.results.cores)
        all_threads = {"count": 0}
        for orig in list(self.graph.centroids):
            i = int(self.graph.nodes_to_indices[orig])
            if i >= self.graph.nodes_to_indices.shape[0]:
                self.report.append(f"Centroid {orig} is beyond the domain of the graph")
            elif self.graph.fs[int(i)] == self.graph.fs[int(i) + 1]:
                self.report.append(f"Centroid {orig} does not exist in the graph")
            else:
                pool.apply_async(self.__func_skim_thread, args=(orig, all_threads))
        pool.close()
        pool.join()
        self.aux_res = None
        self.procedure_id = uuid4().hex
        self.procedure_date = str(datetime.today())

        if pyqt:
            self.skimming.emit(["text skimming", "Saving Outputs"])
            self.skimming.emit(["finished_threaded_procedure", None])

    def save_to_project(self, name: str, format="omx") -> None:
        """Saves skim results to the project folder and creates record in the database

        Args:
            *name* (:obj:`str`): Name of the matrix. Same value for matrix record name and file (plus extension)
            *format* (:obj:`str`, `Optional`): File format ('aem' or 'omx'). Default is 'omx'
        """

        file_name = f"{name}.{format.lower()}"
        mats = Matrices()
        record = mats.new_record(name, file_name, self.results.skims)
        record.procedure_id = self.procedure_id
        record.timestamp = self.procedure_date
        record.procedure = "Network skimming"
        record.save()

    def __func_skim_thread(self, origin, all_threads):
        if threading.get_ident() in all_threads:
            th = all_threads[threading.get_ident()]
        else:
            all_threads[threading.get_ident()] = all_threads["count"]
            th = all_threads["count"]
            all_threads["count"] += 1
        x = skimming_single_origin(origin, self.graph, self.results, self.aux_res, th)
        self.cumulative += 1
        if x != origin:
            self.report.append(x)
        if pyqt:
            self.skimming.emit(["zones finalized", self.cumulative])
            txt = str(self.cumulative) + " / " + str(self.matrix.zones)
            self.skimming.emit(["text skimming", txt])



import os
from sqlite3 import Connection
from aequilibrae.parameters import Parameters
from aequilibrae import logger


def spatialite_connection(conn: Connection) -> Connection:
    conn.enable_load_extension(True)
    par = Parameters()
    spatialite_path = par.parameters["system"]["spatialite_path"]
    if spatialite_path not in os.environ['PATH']:
        os.environ['PATH'] = spatialite_path + ';' + os.environ['PATH']
    try:
        conn.load_extension("mod_spatialite")
    except Exception as e:
        logger.warning(f"AequilibraE might not work as intended without spatialite. {e.args}")
    return conn



import re
from sqlite3 import Connection
from os.path import join, dirname, realpath
from aequilibrae import logger

req_link_flds = ["link_id", "a_node", "b_node", "direction", "distance", "modes", "link_type"]
req_node_flds = ["node_id", "is_centroid"]
protected_fields = ['ogc_fid', 'geometry']


def initialize_tables(conn: Connection) -> None:
    create_base_tables(conn)
    populate_meta_extra_attributes(conn)
    add_triggers(conn)


def create_base_tables(conn: Connection) -> None:
    spec_folder = join(dirname(realpath(__file__)), "database_specification", 'tables')
    with open(join(spec_folder, 'table_list.txt'), 'r') as file_list:
        all_tables = file_list.readlines()
    all_tables = [x.rstrip() for x in all_tables]
    for f in all_tables:
        qry_file = join(spec_folder, f'{f}.sql')
        run_queries_from_sql_file(conn, qry_file)


def populate_meta_extra_attributes(conn: Connection) -> None:
    extra_keys = ['alpha', 'beta', 'gamma', 'delta', 'epsilon', 'zeta', 'iota', 'sigma', 'phi', 'tau']
    extra_keys = [[x, 'Available for user convenience'] for x in extra_keys]

    cursor = conn.cursor()
    for table_name in ['link_types', 'modes']:
        for f, d in extra_keys:
            sql = f"INSERT INTO 'attributes_documentation' (name_table, attribute, description) VALUES('{table_name}','{f}', '{d}')"
            cursor.execute(sql)
    conn.commit()


def add_triggers(conn: Connection) -> None:
    """Adds consistency triggers to the project"""
    spec_folder = join(dirname(realpath(__file__)), "database_specification", 'triggers')
    with open(join(spec_folder, 'triggers_list.txt'), 'r') as file_list:
        all_trigger_sets = file_list.readlines()
    all_trigger_sets = [x.rstrip() for x in all_trigger_sets]
    for f in all_trigger_sets:
        qry_file = join(spec_folder, f'{f}.sql')
        run_queries_from_sql_file(conn, qry_file)


def remove_triggers(conn: Connection) -> None:
    curr = conn.cursor()
    spec_folder = join(dirname(realpath(__file__)), "database_specification", 'triggers')
    with open(join(spec_folder, 'triggers_list.txt'), 'r') as file_list:
        all_trigger_sets = file_list.readlines()

    create_drop_regex = re.compile(r'create\s+trigger\s+(\w+)', flags=re.I)
    for table in all_trigger_sets:
        qry_file = join(spec_folder, f'{table.rstrip()}.sql')

        with open(qry_file, "r") as sql_file:
            query_list = sql_file.read()

        # Running one query/command at a time helps debugging in the case a particular command fails
        for cmd in query_list.split("--#"):
            for qry in cmd.split("\n"):
                if qry[:2] == '--':
                    continue
                while '  ' in qry:
                    qry = qry.replace('  ', ' ')

                m = re.search(create_drop_regex, qry)
                if m:
                    try:
                        curr.execute(f'drop trigger if exists {m.group(1).lower()}')
                    except Exception as e:
                        logger.error(f'Failed removing triggers table - > {e.args}')
                        logger.error(f'Point of failure - > {qry}')
        conn.commit()


def run_queries_from_sql_file(conn: Connection, qry_file: str) -> None:
    curr = conn.cursor()

    with open(qry_file, "r") as sql_file:
        query_list = sql_file.read()

    # Running one query/command at a time helps debugging in the case a particular command fails
    for cmd in query_list.split("--#"):
        try:
            curr.execute(cmd)
        except Exception as e:
            msg = f"Error running SQL command: {e.args}"
            logger.error(msg)
            logger.info(cmd)
            raise e



import os
import sqlite3
import importlib.util as iutil
from aequilibrae.project.spatialite_connection import spatialite_connection

spec = iutil.find_spec("qgis")
inside_qgis = spec is not None
if inside_qgis:
    import qgis

ENVIRON_VAR = "AEQUILIBRAE_PROJECT_PATH"


def database_connection() -> sqlite3.Connection:
    if ENVIRON_VAR not in os.environ:
        raise FileExistsError("There is no AequilibraE project loaded to connect to")

    path = os.environ.get(ENVIRON_VAR)
    file_name = os.path.join(path, "project_database.sqlite")
    if inside_qgis:
        conn = qgis.utils.spatialite_connect(file_name)
    else:
        conn = spatialite_connection(sqlite3.connect(file_name))
    return conn



from unittest import TestCase
import os
import pathlib
import sqlite3
import uuid
import string
import random
from random import choice
from tempfile import gettempdir
import numpy as np
import pandas as pd
from aequilibrae import TrafficAssignment, TrafficClass, Graph
from aequilibrae.utils.create_example import create_example

from ...data import siouxfalls_project


class TestTrafficAssignment(TestCase):
    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(gettempdir(), "temp_data") + ";" + os.environ["PATH"]

        proj_path = os.path.join(gettempdir(), "test_traffic_assignment_" + uuid.uuid4().hex)
        self.project = create_example(proj_path)
        self.project.network.build_graphs()
        self.car_graph = self.project.network.graphs["c"]  # type: Graph
        self.car_graph.set_graph("free_flow_time")
        self.car_graph.set_blocked_centroid_flows(False)

        self.matrix = self.project.matrices.get_matrix("demand_omx")
        self.matrix.computational_view()

        self.assignment = TrafficAssignment()
        self.assigclass = TrafficClass("car", self.car_graph, self.matrix)

        self.algorithms = ["msa", "cfw", "bfw", "frank-wolfe"]

    def tearDown(self) -> None:
        self.matrix.close()
        self.project.close()

    def test_matrix_with_wrong_type(self):
        self.matrix.matrix_view = np.array(self.matrix.matrix_view, np.int32)
        with self.assertRaises(TypeError):
            _ = TrafficClass("car", self.car_graph, self.matrix)

    def test_set_vdf(self):
        with self.assertRaises(ValueError):
            self.assignment.set_vdf("CQS")

        self.assignment.set_vdf("BPR")

    def test_set_classes(self):
        with self.assertRaises(AttributeError):
            self.assignment.set_classes([1, 2])

        with self.assertRaises(Exception):
            self.assignment.set_classes(self.assigclass)

        self.assignment.set_classes([self.assigclass])
        # self.fail()

    def test_algorithms_available(self):
        algs = self.assignment.algorithms_available()
        real = ["all-or-nothing", "msa", "frank-wolfe", "bfw", "cfw", "fw"]

        diff = [x for x in real if x not in algs]
        diff2 = [x for x in algs if x not in real]

        if len(diff) + len(diff2) > 0:
            self.fail("list of algorithms raised is wrong")

    def test_set_cores(self):
        with self.assertRaises(Exception):
            self.assignment.set_cores(3)

        self.assignment.add_class(self.assigclass)
        with self.assertRaises(ValueError):
            self.assignment.set_cores("q")

        self.assignment.set_cores(3)

    def test_set_algorithm(self):
        with self.assertRaises(AttributeError):
            self.assignment.set_algorithm("not an algo")

        self.assignment.add_class(self.assigclass)

        with self.assertRaises(Exception):
            self.assignment.set_algorithm("msa")

        self.assignment.set_vdf("BPR")
        self.assignment.set_vdf_parameters({"alpha": "b", "beta": "power"})

        self.assignment.set_capacity_field("capacity")
        self.assignment.set_time_field("free_flow_time")

        self.assignment.max_iter = 10

        for algo in self.algorithms:
            for _ in range(10):
                algo = "".join([x.upper() if random.random() < 0.5 else x.lower() for x in algo])
                self.assignment.set_algorithm(algo)

        with self.assertRaises(AttributeError):
            self.assignment.set_algorithm("not a valid algorithm")

    def test_set_vdf_parameters(self):
        with self.assertRaises(Exception):
            self.assignment.set_vdf_parameters({"alpha": "b", "beta": "power"})

        self.assignment.set_vdf("bpr")
        self.assignment.add_class(self.assigclass)
        self.assignment.set_vdf_parameters({"alpha": "b", "beta": "power"})

    def test_set_time_field(self):
        with self.assertRaises(ValueError):
            self.assignment.set_time_field("capacity")

        self.assignment.add_class(self.assigclass)

        N = random.randint(1, 50)
        val = "".join(random.choices(string.ascii_uppercase + string.digits, k=N))
        with self.assertRaises(ValueError):
            self.assignment.set_time_field(val)

        self.assignment.set_time_field("free_flow_time")
        self.assertEqual(self.assignment.time_field, "free_flow_time")

    def test_set_capacity_field(self):
        with self.assertRaises(ValueError):
            self.assignment.set_capacity_field("capacity")

        self.assignment.add_class(self.assigclass)

        N = random.randint(1, 50)
        val = "".join(random.choices(string.ascii_uppercase + string.digits, k=N))
        with self.assertRaises(ValueError):
            self.assignment.set_capacity_field(val)

        self.assignment.set_capacity_field("capacity")
        self.assertEqual(self.assignment.capacity_field, "capacity")

    def test_set_save_path_files(self):
        self.assignment.set_classes([self.assigclass])
        # make sure default is false
        for c in self.assignment.classes:
            self.assertEqual(c._aon_results.save_path_file, False)
        self.assignment.set_save_path_files(True)
        for c in self.assignment.classes:
            self.assertEqual(c._aon_results.save_path_file, True)

        # reset for most assignment tests
        self.assignment.set_save_path_files(False)
        for c in self.assignment.classes:
            self.assertEqual(c._aon_results.save_path_file, False)

    def test_set_path_file_format(self):
        self.assignment.set_classes([self.assigclass])
        with self.assertRaises(Exception):
            self.assignment.set_path_file_format("shiny_format")
        self.assignment.set_path_file_format("parquet")
        for c in self.assignment.classes:
            self.assertEqual(c._aon_results.write_feather, False)
        self.assignment.set_path_file_format("feather")
        for c in self.assignment.classes:
            self.assertEqual(c._aon_results.write_feather, True)

    def test_save_path_files(self):
        self.assignment.add_class(self.assigclass)
        self.assignment.set_save_path_files(True)

        self.assignment.set_vdf("BPR")
        self.assignment.set_vdf_parameters({"alpha": 0.15, "beta": 4.0})
        self.assignment.set_vdf_parameters({"alpha": "b", "beta": "power"})

        self.assignment.set_capacity_field("capacity")
        self.assignment.set_time_field("free_flow_time")

        self.assignment.max_iter = 2
        self.assignment.set_algorithm("msa")
        self.assignment.execute()

        path_file_dir = pathlib.Path(self.project.project_base_path) / "path_files" / self.assignment.procedure_id
        self.assertTrue(path_file_dir.is_dir())

        # compare everything to reference files. Note that there is no graph simplification happening in SiouxFalls
        # and therefore we compare the files directly, otherwise a translation from the simplified ids to link_ids
        # would need to be performed.
        # Reference files were generated on 12/6/21, any changes to the test project will need to be applied to the
        # reference files. Also, the name given to the traffic class (see setUp above) has to be "car".
        class_id = f"c{self.assigclass.mode}_{self.assigclass.__id__}"
        reference_path_file_dir = pathlib.Path(siouxfalls_project) / "path_files"

        ref_node_correspondence = pd.read_feather(reference_path_file_dir / f"nodes_to_indeces_{class_id}.feather")
        node_correspondence = pd.read_feather(path_file_dir / f"nodes_to_indeces_{class_id}.feather")
        ref_node_correspondence.node_index = ref_node_correspondence.node_index.astype(node_correspondence.node_index.dtype)
        self.assertTrue(node_correspondence.equals(ref_node_correspondence))

        ref_correspondence = pd.read_feather(reference_path_file_dir / f"correspondence_{class_id}.feather")
        correspondence = pd.read_feather(path_file_dir / f"correspondence_{class_id}.feather")
        for col in correspondence.columns:
            ref_correspondence[col] = ref_correspondence[col].astype(correspondence[col].dtype)
        self.assertTrue(correspondence.equals(ref_correspondence))

        path_class_id = f"path_{class_id}"
        for i in range(1, self.assignment.max_iter + 1):
            class_dir = path_file_dir / f"iter{i}" / path_class_id
            ref_class_dir = reference_path_file_dir / f"iter{i}" / path_class_id
            for o in self.assigclass.matrix.index:
                o_ind = self.assigclass.graph.compact_nodes_to_indices[o]
                this_o_path_file = pd.read_feather(class_dir / f"o{o_ind}.feather")
                ref_this_o_path_file = pd.read_feather(ref_class_dir / f"o{o_ind}.feather")
                is_eq = this_o_path_file == ref_this_o_path_file
                self.assertTrue(is_eq.all().all())

                this_o_index_file = pd.read_feather(class_dir / f"o{o_ind}_indexdata.feather")
                ref_this_o_index_file = pd.read_feather(ref_class_dir / f"o{o_ind}_indexdata.feather")
                is_eq = this_o_index_file == ref_this_o_index_file
                self.assertTrue(is_eq.all().all())

        self.assignment.set_save_path_files(False)

    def test_execute_and_save_results(self):
        conn = sqlite3.connect(os.path.join(siouxfalls_project, "project_database.sqlite"))
        results = pd.read_sql("select volume from links order by link_id", conn)

        self.assignment.add_class(self.assigclass)
        self.assignment.set_vdf("BPR")
        self.assignment.set_vdf_parameters({"alpha": 0.15, "beta": 4.0})
        self.assignment.set_vdf_parameters({"alpha": "b", "beta": "power"})

        self.assignment.set_capacity_field("capacity")
        self.assignment.set_time_field("free_flow_time")

        self.assignment.max_iter = 10
        self.assignment.set_algorithm("msa")
        self.assignment.execute()

        with self.assertRaises(ValueError):
            # We have no skimming setup
            self.assignment.save_skims("my_skims", "all")

        msa10 = self.assignment.assignment.rgap

        correl = np.corrcoef(self.assigclass.results.total_link_loads, results.volume.values)[0, 1]
        self.assertLess(0.8, correl)

        self.assignment.max_iter = 50
        self.assignment.set_algorithm("msa")
        self.assignment.execute()
        msa25 = self.assignment.assignment.rgap

        correl = np.corrcoef(self.assigclass.results.total_link_loads, results.volume)[0, 1]
        self.assertLess(0.98, correl)

        self.assignment.set_algorithm("frank-wolfe")
        self.assignment.execute()

        fw25 = self.assignment.assignment.rgap

        correl = np.corrcoef(self.assigclass.results.total_link_loads, results.volume)[0, 1]
        self.assertLess(0.99, correl)

        self.assignment.set_algorithm("cfw")
        self.assignment.execute()
        cfw25 = self.assignment.assignment.rgap

        correl = np.corrcoef(self.assigclass.results.total_link_loads, results.volume)[0, 1]
        self.assertLess(0.995, correl)

        # For the last algorithm, we set skimming
        self.car_graph.set_skimming(["free_flow_time", "distance"])
        assigclass = TrafficClass("car", self.car_graph, self.matrix)
        self.assignment.set_classes([assigclass])

        self.assignment.set_algorithm("bfw")
        self.assignment.execute()
        bfw25 = self.assignment.assignment.rgap

        correl = np.corrcoef(self.assigclass.results.total_link_loads, results.volume)[0, 1]
        self.assertLess(0.999, correl)

        self.assertLess(msa25, msa10)
        self.assertLess(fw25, msa25)
        self.assertLess(cfw25, fw25)
        self.assertLess(bfw25, cfw25)

        self.assignment.save_results("save_to_database")
        self.assignment.save_skims("my_skims", "all")

        with self.assertRaises(ValueError):
            self.assignment.save_results("save_to_database")

    def test_info(self):
        iterations = random.randint(1, 10000)
        rgap = random.random() / 10000
        algo = choice(self.algorithms)

        self.assignment.add_class(self.assigclass)
        self.assignment.set_vdf("BPR")
        self.assignment.set_vdf_parameters({"alpha": 0.15, "beta": 4.0})
        self.assignment.set_vdf_parameters({"alpha": "b", "beta": "power"})

        self.assignment.set_capacity_field("capacity")
        self.assignment.set_time_field("free_flow_time")

        self.assignment.max_iter = iterations
        self.assignment.rgap_target = rgap
        self.assignment.set_algorithm(algo)

        # TY
        for _ in range(10):
            algo = "".join([x.upper() if random.random() < 0.5 else x.lower() for x in algo])

        dct = self.assignment.info()
        if algo.lower() == "fw":
            algo = "frank-wolfe"
        self.assertEqual(dct["Algorithm"], algo.lower(), "Algorithm not correct in info method")

        self.assertEqual(dct["Maximum iterations"], iterations, "maximum iterations not correct in info method")



import numpy as np


class MultiThreadedAoN:
    def __init__(self):
        # The predecessors for each node in the graph
        self.predecessors = np.array([])
        # holds the skims for all nodes in the network (during path finding)
        self.temporary_skims = np.array([])
        # Keeps the order in which the nodes were reached for the cascading network loading
        self.reached_first = np.array([])
        # The previous link for each node in the tree
        self.connectors = np.array([])
        # Temporary results for assignment. Necessary for parallelization
        self.temp_link_loads = np.array([])
        # Temporary nodes for assignment. Necessary for cascading
        self.temp_node_loads = np.array([])
        #  holds the b_nodes in case of flows through centroid connectors are blocked
        self.temp_b_nodes = np.array([])

    # In case we want to do by hand, we can prepare each method individually
    def prepare(self, graph, results):
        itype = graph.default_types("int")
        ftype = graph.default_types("float")
        self.predecessors = np.zeros((results.compact_nodes, results.cores), dtype=itype)
        if results.num_skims > 0:
            self.temporary_skims = np.zeros((results.compact_nodes, results.num_skims, results.cores), dtype=ftype)
        else:
            self.temporary_skims = np.zeros((1, 1, results.cores), dtype=ftype)
        self.reached_first = np.zeros((results.compact_nodes, results.cores), dtype=itype)
        self.connectors = np.zeros((results.compact_nodes, results.cores), dtype=itype)
        self.temp_link_loads = np.zeros((results.links + 1, results.classes["number"], results.cores), dtype=ftype)
        self.temp_node_loads = np.zeros((results.compact_nodes, results.classes["number"], results.cores), dtype=ftype)
        self.temp_b_nodes = np.zeros((graph.compact_graph.b_node.shape[0], results.cores), dtype=itype)
        for i in range(results.cores):
            self.temp_b_nodes[:, i] = graph.compact_graph.b_node.values[:]



import sys
import os
import platform
import numpy as np
import pyarrow as pa
from setuptools import setup, find_packages
from setuptools import Extension
from Cython.Distutils import build_ext
from aequilibrae.paths.__version__ import release_version

sys.dont_write_bytecode = True

here = os.path.dirname(os.path.realpath(__file__))
whole_path = os.path.join(here, "aequilibrae/paths", "AoN.pyx")
ext_module = Extension(
    "aequilibrae.paths.AoN", [whole_path], include_dirs=[np.get_include(), pa.get_include()], language="c++"
)

# this is for building pyarrow on platforms w/o wheel, like our one of our macos/python combos
if "WINDOWS" not in platform.platform().upper():
    ext_module.extra_compile_args.append("-std=c++11")

pkgs = [pkg for pkg in find_packages()]

pkg_data = {
    "aequilibrae.reference_files": ["spatialite.sqlite", "nauru.zip", "sioux_falls.zip"],
    "aequilibrae.paths": ["parameters.pxi"],
    "aequilibrae": ["parameters.yml"],
    "aequilibrae.project": ["database_specification/tables/*.*", "database_specification/triggers/*.*"],
}
loose_modules = ["__version__", "parameters"]

if __name__ == "__main__":
    setup(
        name="aequilibrae",
        version=release_version,
        # TODO: Fix the requirements and optional requirements to bring directly from the requirements file
        install_requires=["numpy", "pyaml", "pandas", "requests", "shapely", "scipy", "pyarrow"],
        packages=pkgs,
        package_dir={"": "."},
        py_modules=loose_modules,
        package_data=pkg_data,
        zip_safe=False,
        description="A package for transportation modeling",
        author="Pedro Camargo",
        author_email="c@margo.co",
        url="https://github.com/AequilibraE/aequilibrae",
        license="See license.txt",
        classifiers=[
            "Programming Language :: Python",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
        ],
        cmdclass={"build_ext": build_ext},
        ext_modules=[ext_module],
    )



import string
from .safe_class import SafeClass
from aequilibrae.project.database_connection import database_connection


class LinkType(SafeClass):
    """A link_type object represents a single record in the *link_types* table"""
    __alowed_characters = string.ascii_letters + '_'

    def delete(self):
        conn = database_connection()
        curr = conn.cursor()
        curr.execute(f'DELETE FROM link_types where link_type_id="{self.link_type_id}"')
        conn.commit()
        del self

    def save(self):
        conn = database_connection()
        curr = conn.cursor()

        curr.execute(f'select count(*) from link_types where link_type_id="{self.link_type_id}"')
        if curr.fetchone()[0] == 0:
            data = [self.link_type_id, self.link_type]
            curr.execute('Insert into link_types (link_type_id, link_type) values(?,?)', data)

        for key, value in self.__dict__.items():
            if key != 'link_type_id' and key in self.__original__:
                v_old = self.__original__.get(key, None)
                if value != v_old and value is not None:
                    self.__original__[key] = value
                    curr.execute(f"update 'link_types' set '{key}'=? where link_type_id='{self.link_type_id}'", [value])
        conn.commit()
        conn.close()

    def __setattr__(self, instance, value) -> None:
        if instance == 'link_type':
            if isinstance(value, str):
                if not len(value):
                    raise ValueError('link_type cannot be zero-length')
                for letter in value:
                    if letter not in self.__alowed_characters:
                        raise ValueError('link_type can only contain letters and "_"')
            else:
                raise ValueError('link_type must be string')
        if instance == 'link_type_id':
            raise ValueError('Changing a link_type_id is not supported. Create a new one and delete this')
        else:
            self.__dict__[instance] = value



from unittest import TestCase
import tempfile
import os
from aequilibrae.project import Project
from aequilibrae.project.database_connection import database_connection
import uuid


class TestProject(TestCase):
    def setUp(self) -> None:
        self.temp_proj_folder = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
        self.proj = Project()
        self.proj.new(self.temp_proj_folder)

    def tearDown(self) -> None:
        self.proj.close()

    def test_opening_wrong_folder(self):
        temp_proj_folder = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
        self.proj.close()
        with self.assertRaises(FileNotFoundError):
            proj = Project()
            proj.open(temp_proj_folder)
        self.proj.open(self.temp_proj_folder)

    def test_create_when_already_exists(self):
        with self.assertRaises(Exception):
            q = Project()
            q.new(os.path.join(tempfile.gettempdir(), uuid.uuid4().hex))

        with self.assertRaises(Exception):
            q = Project()
            q.open(os.path.join(tempfile.gettempdir(), uuid.uuid4().hex))

    def test_creation(self):

        curr = self.proj.conn.cursor()
        curr.execute("""PRAGMA table_info(links);""")
        fields = curr.fetchall()
        fields = [x[1] for x in fields]

        if 'distance' not in fields:
            self.fail("Table LINKS was not created correctly")

        curr = self.proj.conn.cursor()
        curr.execute("""PRAGMA table_info(nodes);""")
        nfields = curr.fetchall()
        nfields = [x[1] for x in nfields]

        if 'is_centroid' not in nfields:
            self.fail('Table NODES was not created correctly')

    def test_close(self):

        _ = database_connection()

        self.proj.close()
        with self.assertRaises(FileExistsError):
            _ = database_connection()



import os
from shutil import copytree
import uuid
from tempfile import gettempdir
from unittest import TestCase
from ...data import siouxfalls_project
from aequilibrae.project import Project


class TestLog(TestCase):
    def setUp(self) -> None:
        os.environ['PATH'] = os.path.join(gettempdir(), 'temp_data') + ';' + os.environ['PATH']
        self.proj_dir = os.path.join(gettempdir(), uuid.uuid4().hex)
        copytree(siouxfalls_project, self.proj_dir)

        self.project = Project()
        self.project.open(self.proj_dir)

    def tearDown(self) -> None:
        self.project.close()

    def test_contents(self):
        log = self.project.log()
        cont = log.contents()
        self.assertEqual(len(cont), 4, 'Returned the wrong amount of data from the log')

    def test_clear(self):
        log = self.project.log()
        log.clear()

        with open(os.path.join(self.proj_dir, "aequilibrae.log"), 'r') as file:
            q = file.readlines()
        self.assertEqual(len(q), 0, 'Failed to clear the log file')
